﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using System.IO;
using Microsoft.Win32;
using Microsoft.Win32.SafeHandles;
using System.Runtime.InteropServices;

namespace MemoryMappedNative
{
    unsafe public class FileMmByte : IEnumerable<byte[]>, IEnumerable<IndexedLine>
    {
        private const uint PAGE_READONLY = 0x00000002;
        private const uint PAGE_READWRITE = 0x00000004;
        private const uint FILE_MAP_READ = 0x00000004;
        private const uint FILE_MAP_ALL_ACCESS = 0x00000002;

        [DllImport("kernel32.dll", CharSet = CharSet.Auto, SetLastError = true)]
        private static extern SafeFileHandle CreateFile(String lpFileName,
            FileAccess dwDesiredAccess, FileShare dwShareMode,
            IntPtr lpSecurityAttributes, FileMode dwCreationDisposition,
            uint dwFlagsAndAttributes, IntPtr hTemplateFile);

        [DllImport("kernel32.dll", CharSet = CharSet.Auto, SetLastError = true)]
        private static extern bool CloseHandle(SafeFileHandle h);

        [DllImport("kernel32.dll", SetLastError = true)]
        private static extern SafeFileHandle CreateFileMapping(SafeFileHandle hFile,
            IntPtr lpSecurityAttributes, UInt32 dwProtect, UInt32 dwMaximumSizeHigh,
            UInt32 dwMaximumSizeLow, IntPtr lpMapName);

        [DllImport("kernel32.dll", SetLastError = true)]
        private static extern byte* MapViewOfFile(SafeFileHandle hMapObject,
            UInt32 dwAccess, Int32 dwOffsetHigh, UInt32 dwOffsetLow, UInt64 dwNumberOfBytesToMap);

        [DllImport("kernel32.dll", SetLastError = true)]
        private static extern bool UnmapViewOfFile(byte* lpBaseAddress);

        [DllImport("kernel32.dll", SetLastError = true)]
        private static extern int GetLastError();

        /// <summary>
        /// Mmeory Mapped file access constructor.
        /// </summary>
        /// <param name="pathName">Input file path name</param>
        /// <param name="readOnly">read only or read-write?</param>

        public FileMmByte(String pathName, bool readOnly)
        {
            _fileHandle = CreateFile(pathName, FileAccess.Read, readOnly ? FileShare.Read : FileShare.ReadWrite, IntPtr.Zero, FileMode.Open, 0, IntPtr.Zero);
            if (_fileHandle.IsInvalid) throw new Exception("Error. Failed to open file " + pathName + "Error: " + GetLastError().ToString());

            FileInfo fi = new FileInfo(pathName);
            _FileSize = fi.Length;

            SafeFileHandle fileMapHandle = CreateFileMapping(_fileHandle, IntPtr.Zero, readOnly ? PAGE_READONLY : PAGE_READWRITE, 0, 0, IntPtr.Zero);
            if (fileMapHandle == null) throw new Exception("Failed to CreateFileMapping for mapping for file " + pathName + "Error: " + GetLastError().ToString());

            _pFile = (byte*) MapViewOfFile(fileMapHandle, readOnly ? FILE_MAP_READ : FILE_MAP_ALL_ACCESS, 0, 0, (UInt64)0); CloseHandle(fileMapHandle);
            if (_pFile == null) throw new Exception("Error. Failed to map view of file " + pathName + "Error: " + GetLastError().ToString());
        }

        private Int64 _FileSize;
        private byte* _pFile;
        private SafeFileHandle _fileHandle;
        private int _maxFields;

        public Int64 FileSize { get { return _FileSize; } }
        public byte* pFile { get { return _pFile; } }
        public int MaxFields { get { return _maxFields; } set { _maxFields = value; } }

        /// <summary>
        /// IEnumerable<byte[]> support.
        /// To be done. Combine this class and FileMmChar into a single generic class.
        /// </summary>
        /// <returns></returns>
        IEnumerator<byte[]> IEnumerable<byte[]>.GetEnumerator()
        {
            return new ByteLineEnumerator<byte[]>(this);
        }

        IEnumerator<IndexedLine> IEnumerable<IndexedLine>.GetEnumerator()
        {
            return new ByteLineEnumeratorFields<IndexedLine>(this, _maxFields);
        }

        System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
        {
            throw new NotImplementedException();
        }

        /// <summary>
        /// ReadLines version to return a byte[].
        /// </summary>
        /// <param name="path">Name of the file</param>
        /// <returns>A byte[] line. The application must determine field boundaries.</returns>
        public static IEnumerable<byte[]> ReadLines(String path)
        {
            FileMmByte myFile = new FileMmByte(path, true); // Read only
            IEnumerator<byte[]> byteLineEnumerator = new ByteLineEnumerator<byte[]>(myFile);
            foreach (byte[] line in (IEnumerable<byte[]>)myFile)
            {
                yield return line;
            }
            myFile.Dispose();
        }

        /// <summary>ReadLines version to return IndexedLine objects</summary>
        /// <param name="path">File name to be processed</param>
        /// <param name="maxFields">Maximum number of indexable tab delimited fields in a line.</param>
        /// <returns>An IndexedLine consisiting of a byte[] and a unit[] array with the field start positions.</returns>
        /// <remarks>Alert. ReadLines has an additional argument to create an overload for IndexedLine return type
        ///    as the two ReadLines implementations differ only by the return type. Also, the additional argument
        ///    specifies the maximum number of fields.
        /// </remarks>
        public static IEnumerable<IndexedLine> ReadLines(String path, int maxFields)
        {
            FileMmByte myFile = new FileMmByte(path, true); // Read only
            myFile.MaxFields = maxFields;
            IEnumerator<IndexedLine> byteLineEnumerator = new ByteLineEnumeratorFields<IndexedLine>(myFile, maxFields);
            foreach (IndexedLine line in (IEnumerable<IndexedLine>)myFile)
            {
                yield return line;
            }
            myFile.Dispose();
        }


        public void Dispose()
        {
            if (pFile != null) UnmapViewOfFile(pFile);
            if (!_fileHandle.IsInvalid) CloseHandle(_fileHandle);
        }
    }
}
