﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace MemoryMappedNative
{
    unsafe class ByteLineEnumeratorFields<TItem> : IEnumerator<IndexedLine>
    {
        /// <summary>
        /// Supports class FileMmChar IEnumerable<String> interface.
        /// </summary>
        public ByteLineEnumeratorFields(FileMmByte mappedFile, int maxNumberFields)
        {
            this.currentData = mappedFile;
            lineStart = -1;
            currentLength = 0;
            currentBegin = 0;
            maxFields = maxNumberFields;
        }

        private FileMmByte currentData;
        private Int64 lineStart;
        private Int64 currentBegin;
        private int currentLength;
        private uint[] fieldStart;
        private int maxFields;

        IndexedLine IEnumerator<IndexedLine>.Current
        {
            get
            {
                if (this.lineStart < 0 || currentLength <= 0)
                    throw new InvalidOperationException("Use MoveNext before calling Current");
                byte[] line = new byte[this.currentLength];
                Int64 currentEnd = currentBegin + currentLength;
                Int64 i = currentBegin, j = 0;
                while (i < currentEnd)
                {
                    line[j++] = currentData.pFile[i++];
                }
                // ?? - could we use System.Array.Copy() to improve performance?

                return new IndexedLine(line, fieldStart);
            }
        }

        void IDisposable.Dispose()
        {
        }

        object System.Collections.IEnumerator.Current
        {
            get { throw new NotImplementedException(); }
        }

        bool System.Collections.IEnumerator.MoveNext()
        {
            const byte lf = 0x0A, tab = 0x09;
                    
            Int64 fileSizeMinus1 = currentData.FileSize - 1;

            if (lineStart > fileSizeMinus1) return false;
            fieldStart = new uint[maxFields];

            if (this.lineStart < 0) this.lineStart = 0;  // First call to MoveNext
            currentBegin = lineStart;
            fieldStart[0] = 0;
            // Scan the byte[] currentData.pFile[i] from the line starting at file position lineStart and update lineStart.
            // The end of the line is identified by the new line (0x0A) value, not by the byte[] length. Retain the 
            // beginning position. do not move the data; just retain the line boundaries and the field boundaries.
            byte c;
            int iField = 1;
            for (Int64 i = lineStart; i <= fileSizeMinus1; i++)
            {
                c = currentData.pFile[i];
                if (c == tab)
                {
                    if (iField < maxFields) fieldStart[iField] = (uint)(i + 1 - lineStart);
                    iField++;
                }
                else if (c == lf)
                {
                    currentLength = (int)(i - lineStart);
                    lineStart = i + 1; // Prepare for next MoveNext() call
                    break;
                }
            }
            return true;
        }

        void System.Collections.IEnumerator.Reset()
        {
            lineStart = -1;
            currentLength = 0;
        }
    }

    /// <summary>
    /// An indexed line contains two objects: a byte[] line and a uint[] fields which specifies the field start positions on the line.
    /// </summary>
    public class IndexedLine
    {
        private uint[] fields;
        private byte[] line;

        public IndexedLine(byte[] currentLine, uint[] currentFieldStart)
        {
            line = currentLine;
            fields = currentFieldStart;
        }

        public byte[] Line { get { return line; } }
        public uint[] Fields { get { return fields; } }
    }
}
