﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using MemoryMappedNative;

namespace FieldProcessing

{
    /// <summary>
    /// Functions that process tab delimited lines and extract specific field values.
    /// </summary>
    public static class Helper
    {
        const byte lf = 0x0A;
        const byte tab = 0x09;
        const byte zero = 0x30;  // Convert.ToByte(Convert.ToString('0'));

        /// <summary>
        /// Extract a positive integer field from an IndexedLine line. Fields boundaries are in the list associated with the IndexedLine.
        /// </summary>
        /// <param name="line">String value with the current line.</param>
        /// <param name="fieldNumber">Field number (0 based)</param>
        /// <returns>Field value. -1 if there are invalid characters</returns>
        public static int ExtractIntegerField(IndexedLine line, int fieldNumber)
        {
            int value = 0;
            int lineLength = line.Line.Length;
            if (line.Fields.Length >= lineLength || fieldNumber >= line.Fields.Length) return -1;

            uint fieldStart = line.Fields[fieldNumber];
            uint fieldEnd = fieldNumber < line.Fields.Length - 1 ? line.Fields[fieldNumber + 1] - 1 : (uint)lineLength;

            for (uint ib = fieldStart; ib < fieldEnd; ib++)
            {
                byte by = line.Line[ib];

                int digit = (by - zero);
                if (digit >= 0 && digit <= 9)
                {
                    value = 10 * value + digit;
                }
                else
                {   // Invalid character (must be 0-9). Reset the value and quit.
                    value = -1;
                    break;
                }
            }
            return value;
        }

        /// <summary>
        /// Extract a positive integer field from a byte[] line. Fields are tab delimited.
        /// </summary>
        /// <param name="line">String value with the current line.</param>
        /// <param name="fieldNumber">Field number (0 based)</param>
        /// <returns>Field value. -1 if there are invalid characters</returns>
        public static int ExtractIntegerField(byte[] line, int fieldNumber)
        {
            int value = 0, lineLength = line.Length;

            // Skip to the specified field number and extract the decimal value.
            int iField = 0;

            for (int ib = 0; ib < lineLength; ib++)
            {
                byte by = line[ib];
                if (lf == by) { return value; }
                if (tab == by) { iField++; }
                else
                {
                    if (iField == fieldNumber)
                    {
                        int digit = (by - zero);
                        if (digit >= 0 && digit <= 9)
                        {
                            value = 10 * value + digit;
                        }
                        else
                        {   // Invalid character (must be 0-9). Reset the value and quit.
                            value = -1;
                            break;
                        }
                    }
                }
            }
            return value;
        }

        /// <summary>
        /// Extract a positive integer value from a field in a String. Fields are tab delimited.
        /// </summary>
        /// <param name="line">String value with the current line.</param>
        /// <param name="fieldNumber">Field number (0 based)</param>
        /// <returns>Field value. -1 if there are invalid characters.</returns>
        public static int ExtractIntegerField(String line, int fieldNumber)
        {
            int value = 0, iField = 0;
            byte digit;

            // Skip to the specified field number and extract the decimal value.
            foreach (char ch in line)
            {
                if (ch == 0x09) { iField++; if (iField > fieldNumber) break; }
                else
                {
                    if (iField == fieldNumber)
                    {
                        digit = (byte)(ch - 0x30);
                        if (digit >= 0 && digit <= 9)
                        {
                            value = 10 * value + digit;
                        }
                        else
                        {   // Invalid character (must be 0-9). Reset the value and quit.
                            value = -1;
                            break;
                        }
                    }
                }
            }

            return value;
        }
    }
}

