#             Author: JM Hart, Feb 13, 2011. jmhartsoftware.com.  jmhart62@gmail.com, jmh_assoc@hotmail.com.
#
#  Time all geonames variations over a range of Degree of Parallelism values
#    Usage: TimeAll [ DoP [ display ] ]
#              DoP:     Degree of parallelism. Default if field is empty or 0 is the system logical processor count
#              display: Identified records are displayed if this value is non-zero (default if field is empty: 1)
#
#  Currently, there are three variations, ranked by coding ease and convenience:
#    geonamesPLINQ:    This uses the original and improved versions of see Listing 9-2, page 274, of 
#                      "LINQ to Objects Using C# 4.0" by Troy Magennis. Addison-Wesley Microsoft Technology Series, 2010.
#    geonamesParallel: This replaces the PLINQ solutions with a .NET 4.0 Parallel class pattern, the 
#                      "Parallel Loops for Aggregation" pattern described on pages 73-76 of "Parallel Programming with 
#                      Microsoft .NET" (Campbell, Johnson, Miller, and Toub). Microsoft, 2010.
#    geonamesThreads:  This solutions directly creates and manages threads.
#
#    For more information, see my MSDN Magazine article (Jan, 2011): 
#    Data Processing: Parallelism and Performance - http://msdn.microsoft.com/en-us/magazine/gg535673.aspx
#
#  To get the test data file, download and unzip (into this directory) from: 
#       http://download.geonames.org/export/dump/allCountries.zip. 
#  The unzipped file, allCountries.txt (UTF8), is about 825MB.
#
#  Compute the Degree of Parallelism. The default is the number of logical processors
#  Get information about the CPU
$cpu = Get-WmiObject -Class Win32_Processor;
$DoP = $cpu.NumberOfCores;                              # Number of physical processors
$count = $cpu.NumberOfLogicalProcessors;                # Logical processors (for hyper-threading)
Write-Host "CPU Name:", $cpu.Name, ", Description:", $cpu.Description;
Write-Host "Address Width:", $cpu.AddressWidth, ", Clock Speed", $cpu.MaxClockSpeed;
Write-Host "# of processors:", $DoP, ", #Logical processors:", $count;
if ($count -gt 0) { $DoP = $count; }

if ($args.length -ge 1 -and $args[0] -gt 0) { $DoP = $args[0]; }

Write-Host "Maximum Degree of Parallelism: ", $DoP;
$display = 1;
if ($args.length -ge 2) { $display = $args[1]; }

$ThreadingMethods = "MultiThreaded", "ThreadPool";
Write-Host "********";
Write-Host ".NET Threading";
foreach ($method in $ThreadingMethods)
{
    for ($i=1; $i -le $DoP; $i++)
    {
        .\geonamesThreads\bin\x64\Release\geonamesThreads .\allCountries.txt $i $method $display
    }
}

$ParallelMethods = "Original", "Helper", "MMChar", "MMByte", "Index", "IndexFields";
Write-Host "********";
Write-Host "Parallel Class.";
foreach ($method in $ParallelMethods)
{
    for ($i=1; $i -le $DoP; $i++)
    {
        .\geonamesParallel\bin\x64\Release\geonamesParallel .\allCountries.txt $i $method $display
    }
}

Write-Host "********";
Write-Host "PLINQ";
foreach ($method in $ParallelMethods)
{
    for ($i=1; $i -le $DoP; $i++)
    {
        .\geonamesPLINQ\bin\x64\Release\geonamesPLINQ .\allCountries.txt $i $method $display
    }
}