﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Xml.Linq;
using System.Xml;
using System.Data.SqlClient;
using System.Diagnostics;

using System.Linq.Expressions;
using System.Dynamic;
using System.Text;
using System.Collections.Concurrent;
using System.Threading;
using System.Threading.Tasks;

using MemoryMappedNative;
using FieldProcessing;

namespace geonames
{
    /// <summary>
    /// This program is adapted from Listing 9-2, page 274, of "LINQ to Objects Using C# 4.0" by Troy Magennis. Addison-Wesley Microsoft Technology Series, 2010.
    /// Several implementation strategies are compared for performance.
    /// Usage: geonamesPlinq [ file [degreeOfParallelism [ implementation [ display ] ] ] ]
    ///        file: pathname for the geonames database text file. Default: "Data/AllCountries.txt"
    /// 			Download and unzip from: http://download.geonames.org/export/dump/allCountries.zip. The unzipped file, allCountries.txt (UTF8), is about 825MB.
    /// 			The fields are tab-separated, with one geographical location per line.
    /// 			As of Dec 5, 2010, there is at least one incorrect elevation, for "Señal Salumpite  (8600m) - located in PE". This is clearly a mistake, but is not relevant to this example.
    /// 			I can't vouch for any of the data in this file, so don't plan your next vacation based on this information!
    /// 			You can easily test scalability with longer files by concatenating this file to itself.
    ///        degreeOfParallelism: Value to be used with .AsParallel(). Default: Number of processors.
    ///        implementation: Use one of 5 techniques currently available; specify the technique as one of the following strings:
    ///             Original [Default]: As in Listing 9-2, page 274, of "LINQ to Objects Using C# 4.0" by Troy Magennis.
    ///             Helper: Extract the elevation as an integer and the line. Do not extract all the fields as separate strings as is done in the original.
    ///             MMChar: Map the file, returning each line as a String. The FileMmChar class implements the mapping and supports the IEnumerable<String> interface.
    ///                     "Char" since a String is made up of individual 2-byte Char objects.
    ///             MMByte: Map the file, returning each line as a byte[]. The FileMmByte class implements the mapping and supports the IEnumerable<String> interface.
    ///             Index:  Preprocess the file to put the lines in a List{byte[]>. Since lists are idexable, the PLINQ partitioning
    ///                     is much more efficient (it could be Range Partitioning), and the performance is much better.
    ///                     The preprocessing is time consuming, however, so this strategy would only be useful if you can amortize the cost over multiple
    ///                     queries. For convenience, the list index creation uses the FileMmByte class.
    ///             IndexField: An enhancement to "Index" which creates an index to the individual fields within a byte[] line.
    /// 				 POTENTIAL ENHANCEMENT: If the file size is too large compared to physical memory, use the conventional
    /// 					file processing solution. See "Windows Parallelism, Fast File Searching, and Speculative Processing"
    /// 					for more information.
    /// 		display: If 0, don't display the records. If omitted, or non-zero, do display.
    /// </summary>
    class Program
    {
        const int nameColumn = 1;
        const int countryColumn = 8;
        const int elevationColumn = 15;

        static void Main(string[] args)
        {
            String inFile = "Data/AllCountries.txt";
            if (args.Length >= 1) inFile = args[0];

            int degreeOfParallelism = Environment.ProcessorCount;
            if (args.Length >= 2) degreeOfParallelism = int.Parse(args[1]);

            String implementation = "Original";
            if (args.Length >= 3) implementation = args[2];
            Console.WriteLine("GeonamesPLINQ - Data file: {0}. Degree of Parallelism: {1}. Implementation Technique: {2}.", inFile, degreeOfParallelism, implementation);

            bool display = args.Length < 4 || (args.Length >= 4 && int.Parse(args[3]) > 0);

            Stopwatch watch = new Stopwatch();
            watch.Start();
            try
            {
                if (implementation == "Original")
                {
                    ImplementationOriginal(inFile, degreeOfParallelism, display);
                }
                else if (implementation == "Helper")
                {
                    ImplementationHelper(inFile, degreeOfParallelism, display);
                }
                else if (implementation == "MMChar")
                {
                    ImplementationMMChar(inFile, degreeOfParallelism, display);
                }
                else if (implementation == "MMByte")
                {
                    ImplementationMMByte(inFile, degreeOfParallelism, display);
                }
                else if (implementation == "Index")
                {
                    ImplementationIndex(inFile, degreeOfParallelism, display);
                }
                else if (implementation == "IndexFields")
                {
                    ImplementationIndexFields(inFile, degreeOfParallelism, display);
                }
                else
                {
                    Console.WriteLine("Error: Implementation technique {0} is not supported.", implementation);
                }
                Console.WriteLine("Total Elapsed time: {0}.{1} seconds.", watch.ElapsedMilliseconds / 1000, watch.ElapsedMilliseconds % 1000);
            }
            catch (Exception e)
            {
                Console.WriteLine("Implementation Failed. {0}", e.Message);
            }
        }

        /// <summary>
        /// As in Listing 9-2, page 274, of "LINQ to Objects Using C# 4.0" by Troy Magennis.
        /// </summary>
        /// <param name="inFile">Data file path name</param>
        /// <param name="degreeOfParallelism">Used as the argument to the AsParallel.WithDegreeOfParallelism() method</param>
        static void ImplementationOriginal(String inFile, int degreeOfParallelism, bool display)
        {
            var lines = File.ReadLines(Path.Combine(
                Environment.CurrentDirectory, inFile));

            Stopwatch queryWatch = new Stopwatch();
            queryWatch.Start();
            var q = from line in lines.AsParallel().WithDegreeOfParallelism(degreeOfParallelism)
                    let fields = line.Split(new char[] { '\t' })
                    let elevation = string.IsNullOrEmpty(fields[elevationColumn]) ? 0 : int.Parse(fields[elevationColumn])
                    where elevation > 8000 // elevation in m's
                    orderby elevation descending
                    select new
                    {
                        name = fields[nameColumn] ?? "",
                        elevation = elevation,
                        country = fields[countryColumn]
                    };
            Console.WriteLine("Query Count: {0}. Elapsed Query time: {1}.{2} seconds.", q.Count(), queryWatch.ElapsedMilliseconds / 1000, queryWatch.ElapsedMilliseconds % 1000);

            if (display) foreach (var x in q)
            {
                if (x != null)
                {
                    Console.WriteLine("{0} ({1}m) - located in {2}",
                                            x.name, x.elevation, x.country);
                }
            }
        }

        /// <summary>
        /// Similar to Listing 9-2, page 274, of "LINQ to Objects Using C# 4.0" by Troy Magennis, but
        /// Extract the elevation as an integer and the line. Do not extract all the fields as separate strings as is done in the original.
        /// </summary>
        /// <param name="inFile">Data file path name</param>
        /// <param name="degreeOfParallelism">Used as the argument to the AsParallel.WithDegreeOfParallelism() method</param>
        static void ImplementationHelper(String inFile, int degreeOfParallelism, bool display)
        {
            var lines = File.ReadLines(Path.Combine(
                Environment.CurrentDirectory, inFile));

            Stopwatch queryWatch = new Stopwatch();
            queryWatch.Start();
            var q = from line in lines.AsParallel().WithDegreeOfParallelism(degreeOfParallelism)
                    let elevation = Helper.ExtractIntegerField(line, elevationColumn)
                    where elevation > 8000 // elevation in m's
                    orderby elevation descending
                    select new
                    {
                        elevation = elevation,
                        thisLine = line
                    };
            Console.WriteLine("Query Count: {0}. Elapsed Query time: {1}.{2} seconds.", q.Count(), queryWatch.ElapsedMilliseconds / 1000, queryWatch.ElapsedMilliseconds % 1000);

            if (display) foreach (var x in q)
            {
                if (x != null)
                {
                    String[] fields = x.thisLine.Split(new char[] { '\t' });
                    Console.WriteLine("{0} ({1}m) - located in {2}",
                                      fields[nameColumn], fields[elevationColumn], fields[countryColumn]);
                }
            }

        }

        /// <summary>
        /// Map the file, returning each line as a String. The FileMmChar class implements the mapping and supports the IEnumerable<String> interface.
        ///                     "Char" since a String is made up of individual 2-byte Char objects.
        /// </summary>
        /// <param name="inFile">Data file path name</param>
        /// <param name="degreeOfParallelism">Used as the argument to the AsParallel.WithDegreeOfParallelism() method</param>
        static void ImplementationMMChar(String inFile, int degreeOfParallelism, bool display)
        {
            var lines = FileMmChar.ReadLines(Path.Combine(Environment.CurrentDirectory, inFile));

            Stopwatch queryWatch = new Stopwatch();
            queryWatch.Start();
            var q = from line in lines.AsParallel().WithDegreeOfParallelism(degreeOfParallelism)
                    let elevation = Helper.ExtractIntegerField(line, elevationColumn)
                    where elevation > 8000 // elevation in m's
                    orderby elevation descending
                    select new
                    {
                        elevation = elevation,
                        thisLine = line
                    };
            Console.WriteLine("Query Count: {0}. Elapsed Query time: {1}.{2} seconds.", q.Count(), queryWatch.ElapsedMilliseconds / 1000, queryWatch.ElapsedMilliseconds % 1000);

            int foundLine = 0;
            // A fix-up is needed hare. The original file is UTF-8. The following code does not properly process multi-Char symbols, so
            // some location names do not display properly.
            if (display) foreach (var x in q)
            {
                if (x != null)
                {
                    foundLine++;
                    String[] fields = x.thisLine.Split(new char[] { '\t' });
                    Console.WriteLine("{0} ({1}m) - located in {2}",
                        fields[nameColumn], fields[elevationColumn], fields[countryColumn]);
                }
            }
        }

        /// <summary>
        /// Map the file, returning each line as a byte[]. The FileMmByte class implements the mapping and supports the IEnumerable<byte[]> interface.
        /// </summary>
        /// <param name="inFile">Data file path name</param>
        /// <param name="degreeOfParallelism">Used as the argument to the AsParallel.WithDegreeOfParallelism() method</param>
        static void ImplementationMMByte(String inFile, int degreeOfParallelism, bool display)
        {
            var lines = FileMmByte.ReadLines(Path.Combine(Environment.CurrentDirectory, inFile));

            Stopwatch queryWatch = new Stopwatch();
            queryWatch.Start();
            var q = from line in lines.AsParallel().WithDegreeOfParallelism(degreeOfParallelism)
                    let elevation = Helper.ExtractIntegerField(line, elevationColumn)
                    where elevation > 8000 // elevation in m's
                    orderby elevation descending
                    select new
                    {
                        elevation = elevation,
                        thisLine = line
                    };
            int foundLine = 0;
            Console.WriteLine("Query Count: {0}. Elapsed Query time: {1}.{2} seconds.", q.Count(), queryWatch.ElapsedMilliseconds / 1000, queryWatch.ElapsedMilliseconds % 1000);

            System.Text.Encoding encoding = System.Text.UnicodeEncoding.UTF8;
            if (display) foreach (var x in q)
            {
                if (x != null)
                {
                    foundLine++;
                    String line = encoding.GetString(x.thisLine);
                    String[] fields = line.Split(new char[] { '\t' });
                    Console.WriteLine("{0} ({1}m) - located in {2}",
                          fields[nameColumn], fields[elevationColumn], fields[countryColumn]);
                }
            }
        }

        /// <summary>
        /// Preprocess the file to put the lines in a List<byte[]>. Since lists are idexable, the PLINQ partitioning
        ///     is much more efficient (it could be Range Partitioning), and the performance is much better.
        ///     The preprocessing is time consuming, however, so this strategy would only be useful if you can amortize the cost over multiple
        ///     queries. For convenience, the list index creation uses the FileMmByte class.
        /// </summary>
        /// <param name="inFile">Data file path name</param>
        /// <param name="degreeOfParallelism">Used as the argument to the AsParallel.WithDegreeOfParallelism() method</param>
        static void ImplementationIndex(String inFile, int degreeOfParallelism, bool display)
        {
            List<byte[]> lineListByte = new List<byte[]>();

            Stopwatch listWatch = new Stopwatch();
            listWatch.Start();

            var lines = FileMmByte.ReadLines(Path.Combine(Environment.CurrentDirectory, inFile));

            try
            {
                /* This attempt to parallelize the indexing is slower than creating the line list serially.
                var q = from line in lines.AsParallel().WithDegreeOfParallelism(degreeOfParallelism)
                        select new { thisLine = line };
                foreach (var x in q) { if (x != null) { lineListByte.Add(x.thisLine); } }
                */
                foreach (byte[] line in lines) { lineListByte.Add(line); }
            }
            catch (Exception e)
            {
                Console.WriteLine("MakeLineListByte failed. {0}", e.Message);
            }

            Console.WriteLine("Line Count in line list: {0}. Pre-processing time {1}.{2} seconds.", lineListByte.Count, listWatch.ElapsedMilliseconds / 1000, listWatch.ElapsedMilliseconds % 1000);

            Stopwatch queryWatch = new Stopwatch();
            queryWatch.Start();
            var q = from line in lineListByte.AsParallel().WithDegreeOfParallelism(degreeOfParallelism)
                    let elevation = Helper.ExtractIntegerField(line, elevationColumn)
                    where elevation > 8000 // elevation in m's
                    orderby elevation descending
                    select new
                    {
                        elevation = elevation,
                        thisLine = line
                    };
            Console.WriteLine("Query Count: {0}. Elapsed Query time: {1}.{2} seconds.", q.Count(), queryWatch.ElapsedMilliseconds / 1000, queryWatch.ElapsedMilliseconds % 1000);

            int foundLine = 0;
            System.Text.Encoding encoding = System.Text.UnicodeEncoding.UTF8;
            if (display) foreach (var x in q)
            {
                if (x != null)
                {
                    foundLine++;
                    String line = encoding.GetString(x.thisLine);
                    String[] fields = line.Split(new char[] { '\t' });
                    Console.WriteLine("{0} ({1}m) - located in {2}",
                          fields[nameColumn], fields[elevationColumn], fields[countryColumn]);
                }
            }
        }

        /// <summary>
        /// Preprocess the file to put the lines in a List<IndexedLines>. Since lists are idexable, the PLINQ partitioning
        ///     is much more efficient (it could be Range Partitioning), and the performance is much better.
        ///     The preprocessing is time consuming, however, so this strategy would only be useful if you can amortize the cost over multiple
        ///     queries. For convenience, the list index creation uses the FileMmByte class.
        /// </summary>
        /// <param name="inFile">Data file path name</param>
        /// <param name="degreeOfParallelism">Used as the argument to the AsParallel.WithDegreeOfParallelism() method</param>
        /// <remarks>This implementation enhances ImplementationIndex as LineEnumeratorByteFields returns an IndexedLines object,
        /// which contains both a byte[] representing the lines and a uint[] array containing the field boundaries. This eliminates
        /// the need for the helper ExtractIntegerField() function to scan the line for the requested field.
        /// This enhancement should also simplify more general queries than the simple elevation query in the benchmark.
        /// Thanks to Andrew Greenwald for suggesting this idea.
        /// </remarks>
        const int maxFields = 20;

        static void ImplementationIndexFields(String inFile, int degreeOfParallelism, bool display)
        {
            List<IndexedLine> indexedLineList = new List<IndexedLine>();
            // Alert. Note the second argument to ReadLines to convey semantic information about the number of fields allowed
            // in a line. Also, the second argument is required to create an overload as this method returns and IndexdLine, unlike
            // FileMmByte.ReadLines(path) which returns a byte[].
            var lines = FileMmByte.ReadLines(Path.Combine(Environment.CurrentDirectory, inFile), maxFields);

            Stopwatch listWatch = new Stopwatch();
            listWatch.Start();

            try
            {
                foreach (IndexedLine line in lines) { indexedLineList.Add(line); }
            }
            catch (Exception e)
            {
                Console.WriteLine("MakeLineListByte failed. {0}", e.Message);
            }

            // Convert the list to an array to improve access speed. This adds to the inialization overhead, but would be 
            // worthwhile if there are numerous queries. The performance impact, however, is minimal.
            IndexedLine[] lineArray = new IndexedLine[indexedLineList.Count];
            for (int i = 0; i < indexedLineList.Count; i++) lineArray[i] = indexedLineList[i];

            Console.WriteLine("Line Count in line list: {0}. Pre-processing time {1}.{2} seconds.", indexedLineList.Count, listWatch.ElapsedMilliseconds / 1000, listWatch.ElapsedMilliseconds % 1000);

            Stopwatch queryWatch = new Stopwatch();
            queryWatch.Start();
            var q = from line in lineArray.AsParallel().WithDegreeOfParallelism(degreeOfParallelism)
                    let elevation = Helper.ExtractIntegerField(line, elevationColumn)
                    where elevation > 8000 // elevation in m's
                    orderby elevation descending
                    select new
                    {
                        elevation = elevation,
                        thisLine = line
                    };
            Console.WriteLine("Query Count: {0}. Elapsed Query time: {1}.{2} seconds.", q.Count(), queryWatch.ElapsedMilliseconds / 1000, queryWatch.ElapsedMilliseconds % 1000);

            int foundLine = 0;
            System.Text.Encoding encoding = System.Text.UnicodeEncoding.UTF8;
            if (display) foreach (var x in q)
            {
                if (x != null)
                {
                    foundLine++;
                    String line = encoding.GetString(x.thisLine.Line);
                    String[] fields = line.Split(new char[] { '\t' });
                    Console.WriteLine("{0} ({1}m) - located in {2}",
                          fields[nameColumn], fields[elevationColumn], fields[countryColumn]);
                }
            }
        }
    }

}
