﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Xml.Linq;
using System.Xml;
using System.Data.SqlClient;
using System.Diagnostics;

using System.Linq.Expressions;
using System.Dynamic;
using System.Text;
using System.Collections.Concurrent;
using System.Threading;
using System.Threading.Tasks;

using MemoryMappedNative;
using FieldProcessing;

namespace geonames
{
    /// <summary>
    /// Author: Johnson (John) M. Hart. Date Started: January 25, 2011
    /// This program solves the "Geonames" problem using .NET 4.0's Parallel class in order to compare performance and programming effort with the PLINQ and multi-threaded solutions.
    ///     1.   For a description of the Geonames problem, see Listing 9-2, page 274, of "LINQ to Objects Using C# 4.0" by Troy Magennis. Addison-Wesley Microsoft Technology Series, 2010.
    ///     2.   For my performance improvements to the Magennis solution, see the geonamesPLINQ project and my January, 2011 MSDN article, "Data Processing: Parallelism and Performance"
    ///          http://msdn.microsoft.com/en-us/magazine/gg535673.aspx.
    /// Several implementation strategies are compared for performance.
    /// The basic pattern is the "Parallel Loops for Aggregation" pattern described on pages 73-76 of "Parallel Programming with Microsoft .NET" (Campbell, Johnson, Miller, and Toub). Microsoft, 2010.
    /// Alternatively, see pp 67ff http://www.microsoft.com/downloads/en/details.aspx?FamilyID=86b3d32b-ad26-4bb8-a3ae-c1637026c3ee&displaylang=en ("Patterns for Parallel Programming: Understanding and Applying 
    /// Parallel Patterns with the .NET Framework 4") by Stephen Toub.
    /// 
    /// Usage: geonamesPlinq [ file [degreeOfParallelism [ implementation [ display ] ] ] ]
    ///        file: pathname for the geonames database text file. Default: "Data/AllCountries.txt"
    /// 			Download and unzip from: http://download.geonames.org/export/dump/allCountries.zip. The unzipped file, allCountries.txt (UTF8), is about 825MB.
    /// 			The fields are tab-separated, with one geographical location per line.
    /// 			As of Dec 5, 2010, there is at least one incorrect elevation, for "Señal Salumpite  (8600m) - located in PE". This is clearly a mistake, but is not relevant to this example.
    /// 			I can't vouch for any of the data in this file, so don't plan your next vacation based on this information!
    /// 			You can easily test scalability with longer files by concatenating this file to itself.
    ///        degreeOfParallelism: MaxDegreeOfParallelism property to be used with Parallel.ForEach(). Default: Number of processors.
    ///        implementation: Use one of 5 techniques currently available (they are comparable to the 5 PLINQ solutions); specify the technique as one of the following strings:
    ///        
    ///             Original [Default]: A direct conversion of Listing 9-2, page 274, of "LINQ to Objects Using C# 4.0" by Troy Magennis.
    ///                     The conversion uses Parallel.ForEach and aggregation. The lines to process are obtained from File.ReadLines, which is asynchronous.
    ///             Helper: Extract the elevation as an integer and the line. Do not extract all the fields as separate strings as is done in the original.
    ///         The following solutions are similar, differing almost exclusively by the class used that implements ReadLines.
    ///
    ///         The ease of the conversion, together with the performance gains. illustrate some of C#'s power. Lambdas and implicit typing are two important features.
    ///             MMChar: Map the file, returning each line as a String. The FileMmChar class implements the mapping and supports the IEnumerable<String> interface.
    ///                     "Char" since a String is made up of individual 2-byte Char objects.
    ///             MMByte: Map the file, returning each line as a byte[]. The FileMmByte class implements the mapping and supports the IEnumerable<String> interface.
    ///             Index:  Preprocess the file to put the lines in a List{byte[]>. Since lists are idexable, the PLINQ partitioning
    ///                     is much more efficient (it could be Range Partitioning), and the performance is much better.
    ///                     The preprocessing is time consuming, however, so this strategy would only be useful if you can amortize the cost over multiple
    ///                     queries. For convenience, the list index creation uses the FileMmByte class.
    ///             IndexField: An enhancement to "Index" which creates an index to the individual fields within a byte[] line.
    /// 				 POTENTIAL ENHANCEMENT: If the file size is too large compared to physical memory, use the conventional
    /// 					file processing solution. See "Windows Parallelism, Fast File Searching, and Speculative Processing"
    /// 					for more information.
    /// 		display: If 0, don't display the records. If omitted, or non-zero, do display.
    /// Summary: These Parallel solutions are faster than the PLINQ solutions. I'm starting to write an article (Jan 26, 20111). Watch this space.
    /// </summary>
    class Program
    {
        const int nameColumn = 1;
        const int countryColumn = 8;
        const int elevationColumn = 15;

        static void Main(string[] args)
        {
            String inFile = "Data/AllCountries.txt";
            if (args.Length >= 1) inFile = args[0];

            int degreeOfParallelism = Environment.ProcessorCount;
            if (args.Length >= 2) degreeOfParallelism = int.Parse(args[1]);

            String implementation = "Original";
            if (args.Length >= 3) implementation = args[2];
            Console.WriteLine("GeonamesParallel - Data file: {0}. Degree of Parallelism: {1}. Implementation Technique: {2}.", inFile, degreeOfParallelism, implementation);

            bool display = args.Length < 4 || (args.Length >= 4 && int.Parse(args[3]) > 0);

            Stopwatch watch = new Stopwatch();
            watch.Start();
            try
            {
                if (implementation == "Original")
                {
                    ImplementationOriginal(inFile, degreeOfParallelism, display);
                }
                else if (implementation == "Helper")
                {
                    ImplementationHelper(inFile, degreeOfParallelism, display);
                }
                else if (implementation == "MMChar")
                {
                    ImplementationMMChar(inFile, degreeOfParallelism, display);
                }
                else if (implementation == "MMByte")
                {
                    ImplementationMMByte(inFile, degreeOfParallelism, display);
                }
                else if (implementation == "Index")
                {
                    ImplementationIndex(inFile, degreeOfParallelism, display);
                }
                else if (implementation == "IndexFields")
                {
                    ImplementationIndexFields(inFile, degreeOfParallelism, display);
                }
                else
                {
                    Console.WriteLine("Error: Implementation technique {0} is not supported.", implementation);
                }
                Console.WriteLine("Total Elapsed time: {0}.{1} seconds.", watch.ElapsedMilliseconds / 1000, watch.ElapsedMilliseconds % 1000);
            }
            catch (Exception e)
            {
                Console.WriteLine("Implementation Failed. {0}", e.Message);
            }
        }

        /// <summary>
        /// As in Listing 9-2, page 274, of "LINQ to Objects Using C# 4.0" by Troy Magennis.
        /// </summary>
        /// <param name="inFile">Data file path name</param>
        /// <param name="degreeOfParallelism">Used as the argument to the AsParallel.WithDegreeOfParallelism() method</param>
        static void ImplementationOriginal(String inFile, int degreeOfParallelism, bool display)
        {
            var lines = File.ReadLines(Path.Combine(
                Environment.CurrentDirectory, inFile));

            object lockObject = new object();
            List<String> totalFoundList = new List<String>();

            var options = new ParallelOptions() { MaxDegreeOfParallelism = degreeOfParallelism };

            Stopwatch listWatch = new Stopwatch();
            listWatch.Start();
            
            Parallel.ForEach(
                lines,
                options,
                () => new List<String>(),
                (line, loopState, taskFoundList) =>
                {
                    var fields = line.Split(new char[] { '\t' });
                    var elevation = string.IsNullOrEmpty(fields[elevationColumn]) ? 0 : int.Parse(fields[elevationColumn]);
                    if (elevation > 8000) taskFoundList.Add(line);
                    return taskFoundList;
                },

                (localFoundList) =>
                {
                    lock (lockObject)
                    {
                        totalFoundList.AddRange(localFoundList);
                    }
                }
            );

            Console.WriteLine("Query Count: {0}. Elapsed Query time: {1}.{2} seconds.", totalFoundList.Count(), listWatch.ElapsedMilliseconds / 1000, listWatch.ElapsedMilliseconds % 1000);
            totalFoundList.Sort((line1, line2) =>
            {
                if (Helper.ExtractIntegerField(line1, elevationColumn) > Helper.ExtractIntegerField(line2, elevationColumn)) return -1; // Reverse sort
                if (Helper.ExtractIntegerField(line1, elevationColumn) < Helper.ExtractIntegerField(line2, elevationColumn)) return +1;
                return 0;
            });

            if (display) foreach (var x in totalFoundList)
            {
                if (x != null)
                {
                    var fields = x.Split(new char[] { '\t' });
                    Console.WriteLine("{0} ({1}m) - located in {2}",
                                            fields[nameColumn], fields[elevationColumn], fields[countryColumn]);
                }
            }
        }

        /// <summary>
        /// Similar to Listing 9-2, page 274, of "LINQ to Objects Using C# 4.0" by Troy Magennis, but
        /// Extract the elevation as an integer and the line. Do not extract all the fields as separate strings as is done in the original.
        /// </summary>
        /// <param name="inFile">Data file path name</param>
        /// <param name="degreeOfParallelism">Used as the argument to the AsParallel.WithDegreeOfParallelism() method</param>
        static void ImplementationHelper(String inFile, int degreeOfParallelism, bool display)
        {
            var lines = File.ReadLines(Path.Combine(
                Environment.CurrentDirectory, inFile));

            object lockObject = new object();
            List<String> totalFoundList = new List<String>();

            var options = new ParallelOptions() { MaxDegreeOfParallelism = degreeOfParallelism };

            Stopwatch listWatch = new Stopwatch();
            listWatch.Start();
            
            Parallel.ForEach(
                lines,
                options,
                () => new List<String>(),
                (line, loopState, taskFoundList) =>
                {
                    var elevation = Helper.ExtractIntegerField(line, elevationColumn);
                    if (elevation > 8000) taskFoundList.Add(line);
                    return taskFoundList;
                },

                (localFoundList) =>
                {
                    lock (lockObject)
                    {
                        totalFoundList.AddRange(localFoundList);
                    }
                }
            );

            Console.WriteLine("Query Count: {0}. Elapsed Query time: {1}.{2} seconds.", totalFoundList.Count(), listWatch.ElapsedMilliseconds / 1000, listWatch.ElapsedMilliseconds % 1000);
            totalFoundList.Sort((line1, line2) =>
            {
                if (Helper.ExtractIntegerField(line1, elevationColumn) > Helper.ExtractIntegerField(line2, elevationColumn)) return -1; // Reverse sort
                if (Helper.ExtractIntegerField(line1, elevationColumn) < Helper.ExtractIntegerField(line2, elevationColumn)) return +1;
                return 0;
            });

            if (display) foreach (var x in totalFoundList)
            {
                if (x != null)
                {
                    String[] fields = x.Split(new char[] { '\t' });
                    Console.WriteLine("{0} ({1}m) - located in {2}",
                                      fields[nameColumn], fields[elevationColumn], fields[countryColumn]);
                }
            }

        }

        /// <summary>
        /// Map the file, returning each line as a String. The FileMmChar class implements the mapping and supports the IEnumerable<String> interface.
        ///                     "Char" since a String is made up of individual 2-byte Char objects.
        /// </summary>
        /// <param name="inFile">Data file path name</param>
        /// <param name="degreeOfParallelism">Used as the argument to new ParallelOptions() </param>
        static void ImplementationMMChar(String inFile, int degreeOfParallelism, bool display)
        {
            var lines = FileMmChar.ReadLines(Path.Combine(Environment.CurrentDirectory, inFile));

            object lockObject = new object();
            List<String> totalFoundList = new List<String>();

            var options = new ParallelOptions() { MaxDegreeOfParallelism = degreeOfParallelism };

            Stopwatch listWatch = new Stopwatch();
            listWatch.Start();
            
            Parallel.ForEach(
                lines,
                options,
                () => new List<String>(),
                (line, loopState, taskFoundList) =>
                {
                    var elevation = Helper.ExtractIntegerField(line, elevationColumn);
                    if (elevation > 8000) taskFoundList.Add(line);
                    return taskFoundList;
                },

                (localFoundList) =>
                {
                    lock (lockObject)
                    {
                        totalFoundList.AddRange(localFoundList);
                    }
                }
            );

            Console.WriteLine("Query Count: {0}. Elapsed Query time: {1}.{2} seconds.", totalFoundList.Count(), listWatch.ElapsedMilliseconds / 1000, listWatch.ElapsedMilliseconds % 1000);
            totalFoundList.Sort((line1, line2) =>
            {
                if (Helper.ExtractIntegerField(line1, elevationColumn) > Helper.ExtractIntegerField(line2, elevationColumn)) return -1; // Reverse sort
                if (Helper.ExtractIntegerField(line1, elevationColumn) < Helper.ExtractIntegerField(line2, elevationColumn)) return +1;
                return 0;
            });
            
            // A fix-up is needed hare. The original file is UTF-8. The following code does not properly process multi-Char symbols, so
            // some location names do not display properly.
            if (display) foreach (var x in totalFoundList)
            {
                if (x != null)
                {
                    String[] fields = x.Split(new char[] { '\t' });
                    Console.WriteLine("{0} ({1}m) - located in {2}",
                        fields[nameColumn], fields[elevationColumn], fields[countryColumn]);
                }
            }
        }

        /// <summary>
        /// Map the file, returning each line as a byte[]. The FileMmByte class implements the mapping and supports the IEnumerable<byte[]> interface.
        /// </summary>
        /// <param name="inFile">Data file path name</param>
        /// <param name="degreeOfParallelism">Used as the argument to new ParallelOptions() </param>
        static void ImplementationMMByte(String inFile, int degreeOfParallelism, bool display)
        {
            var lines = FileMmByte.ReadLines(Path.Combine(Environment.CurrentDirectory, inFile));

            object lockObject = new object();
            List<byte[]> totalFoundList = new List<byte[]>();

            var options = new ParallelOptions() { MaxDegreeOfParallelism = degreeOfParallelism };

            Stopwatch listWatch = new Stopwatch();
            listWatch.Start();

            Parallel.ForEach(
                lines,
                options,
                () => new List<byte[]>(),
                (line, loopState, taskFoundList) =>
                {
                    var elevation = Helper.ExtractIntegerField(line, elevationColumn);
                    if (elevation > 8000) taskFoundList.Add(line);
                    return taskFoundList;
                },

                (localFoundList) =>
                {
                    lock (lockObject)
                    {
                        totalFoundList.AddRange(localFoundList);
                    }
                }
            );

            Console.WriteLine("Query Count: {0}. Elapsed Query time: {1}.{2} seconds.", totalFoundList.Count(), listWatch.ElapsedMilliseconds / 1000, listWatch.ElapsedMilliseconds % 1000);
            totalFoundList.Sort((line1, line2) =>
            {
                if (Helper.ExtractIntegerField(line1, elevationColumn) > Helper.ExtractIntegerField(line2, elevationColumn)) return -1; // Reverse sort
                if (Helper.ExtractIntegerField(line1, elevationColumn) < Helper.ExtractIntegerField(line2, elevationColumn)) return +1;
                return 0;
            });
            System.Text.Encoding encoding = System.Text.UnicodeEncoding.UTF8;
            if (display) foreach (var x in totalFoundList)
            {
                if (x != null)
                {
                    String line = encoding.GetString(x);
                    String[] fields = line.Split(new char[] { '\t' });
                    Console.WriteLine("{0} ({1}m) - located in {2}",
                          fields[nameColumn], fields[elevationColumn], fields[countryColumn]);
                }
            }
        }

        /// <summary>
        /// Preprocess the file to put the lines in a List<byte[]>. Since lists are indexable, the Parallel.ForEach partitioning
        ///     is much more efficient, and the performance is much better than for a collection that is only IEnumerable.
        ///     The preprocessing is time consuming, however, so this strategy would only be useful if you can amortize the cost over multiple
        ///     queries. For convenience, the list index creation uses the FileMmByte class.
        /// </summary>
        /// <param name="inFile">Data file path name</param>
        /// <param name="degreeOfParallelism">Used as the argument to new ParallelOptions() </param>
        static void ImplementationIndex(String inFile, int degreeOfParallelism, bool display)
        {
            List<byte[]> lineListByte = new List<byte[]>();
            var lines = FileMmByte.ReadLines(Path.Combine(Environment.CurrentDirectory, inFile));

            Stopwatch preprocWatch = new Stopwatch();
            preprocWatch.Start();

            try
            {
                foreach (byte[] line in lines) { lineListByte.Add(line); }
            }
            catch (Exception e)
            {
                Console.WriteLine("MakeLineListByte failed. {0}", e.Message);
            }

            Console.WriteLine("Line Count in line list: {0}. Pre-processing time {1}.{2} seconds.", lineListByte.Count, preprocWatch.ElapsedMilliseconds / 1000, preprocWatch.ElapsedMilliseconds % 1000);

            var options = new ParallelOptions() { MaxDegreeOfParallelism = degreeOfParallelism };

            object lockObject = new object();
            List<byte[]> totalFoundList = new List<byte[]>();

            Stopwatch listWatch = new Stopwatch();
            listWatch.Start();

            Parallel.ForEach(
                lineListByte,
                options,
                () => new List<byte[]>(),
                (line, loopState, taskFoundList) =>
                {
                    var elevation = Helper.ExtractIntegerField(line, elevationColumn);
                    if (elevation > 8000) taskFoundList.Add(line);
                    return taskFoundList;
                },

                (localFoundList) =>
                {
                    lock (lockObject)
                    {
                        totalFoundList.AddRange(localFoundList);
                    }
                }
            );

            Console.WriteLine("Query Count: {0}. Elapsed Query time: {1}.{2} seconds.", totalFoundList.Count(), listWatch.ElapsedMilliseconds / 1000, listWatch.ElapsedMilliseconds % 1000); 

            totalFoundList.Sort((line1, line2) =>
            {
                if (Helper.ExtractIntegerField(line1, elevationColumn) > Helper.ExtractIntegerField(line2, elevationColumn)) return -1; // Reverse sort
                if (Helper.ExtractIntegerField(line1, elevationColumn) < Helper.ExtractIntegerField(line2, elevationColumn)) return +1;
                return 0;
            });

            System.Text.Encoding encoding = System.Text.UnicodeEncoding.UTF8;
            if (display) foreach (var x in totalFoundList)
            {
                if (x != null)
                {
                    String line = encoding.GetString(x);
                    String[] fields = line.Split(new char[] { '\t' });
                    Console.WriteLine("{0} ({1}m) - located in {2}",
                          fields[nameColumn], fields[elevationColumn], fields[countryColumn]);
                }
            }
        }

        /// <summary>
        /// Preprocess the file to put the lines in a List<IndexedLines>. Since lists are indexable, the Parallel.ForEach partitioning
        ///     is much more efficient, and the performance is much better than for a collection that is only IEnumerable.
        ///     The preprocessing is time consuming, however, so this strategy would only be useful if you can amortize the cost over multiple
        ///     queries. For convenience, the list index creation uses the FileMmByte class.
        /// </summary>
        /// <param name="inFile">Data file path name</param>
        /// <param name="degreeOfParallelism">Used as the argument to new ParallelOptions</param>
        /// <remarks>This implementation enhances ImplementationIndex as LineEnumeratorByteFields returns an IndexedLines object,
        /// which contains both a byte[] representing the lines and a uint[] array containing the field boundaries. This eliminates
        /// the need for the helper ExtractIntegerField() function to scan the line for the requested field.
        /// This enhancement should also simplify more general queries than the simple elevation query in the benchmark.
        /// </remarks>
        const int maxFields = 20;

        static void ImplementationIndexFields(String inFile, int degreeOfParallelism, bool display)
        {
            List<IndexedLine> indexedLineList = new List<IndexedLine>();
            // Alert. Note the second argument to ReadLines to convey semantic information about the number of fields allowed
            // in a line. Also, the second argument is required to create an overload as this method returns an IndexdLine, unlike
            // FileMmByte.ReadLines(path) which returns a byte[].
            var lines = FileMmByte.ReadLines(Path.Combine(Environment.CurrentDirectory, inFile), maxFields);

            Stopwatch preprocWatch = new Stopwatch();
            preprocWatch.Start();

            try
            {
                foreach (IndexedLine line in lines) { indexedLineList.Add(line); }
            }
            catch (Exception e)
            {
                Console.WriteLine("MakeLineListByte failed. {0}", e.Message);
            }

            var options = new ParallelOptions() { MaxDegreeOfParallelism = degreeOfParallelism };
            // Convert the list to an array to improve access speed. This adds to the inialization overhead, but would be 
            // worthwhile if there are numerous queries. The performance impact, however, is minimal.
            // the Parallel.For has a small performance impact but is an additonal example.
            IndexedLine[] lineArray = new IndexedLine[indexedLineList.Count];
            Parallel.For(0, indexedLineList.Count, options, (i) => { lineArray[i] = indexedLineList[i]; });

            Console.WriteLine("Line Count in line list: {0}. Pre-processing time {1}.{2} seconds.", indexedLineList.Count, preprocWatch.ElapsedMilliseconds / 1000, preprocWatch.ElapsedMilliseconds % 1000);

            object lockObject = new object();
            List<IndexedLine> totalFoundList = new List<IndexedLine>();

            Stopwatch listWatch = new Stopwatch();
            listWatch.Start(); 
            
            Parallel.ForEach(
                lineArray, 
                options,
                () => new List<IndexedLine>(),
                (line, loopState, taskFoundList) => 
                {
                    var elevation = Helper.ExtractIntegerField(line, elevationColumn);
                    if (elevation > 8000) taskFoundList.Add(line);
                    return taskFoundList;
                },

                (localFoundList) =>
                {
                    lock(lockObject)
                    {
                        totalFoundList.AddRange(localFoundList);
                    }
                } 
            );
            Console.WriteLine("Query Count: {0}. Elapsed Query time: {1}.{2} seconds.", totalFoundList.Count(), listWatch.ElapsedMilliseconds / 1000, listWatch.ElapsedMilliseconds % 1000);

            totalFoundList.Sort((line1, line2) =>
            {
                if (Helper.ExtractIntegerField(line1, elevationColumn) > Helper.ExtractIntegerField(line2, elevationColumn)) return -1; // Reverse sort
                if (Helper.ExtractIntegerField(line1, elevationColumn) < Helper.ExtractIntegerField(line2, elevationColumn)) return +1;
                return 0;
            });

            System.Text.Encoding encoding = System.Text.UnicodeEncoding.UTF8;
            if (display) foreach (var x in totalFoundList)
            {
                if (x != null)
                {
                    String line = encoding.GetString(x.Line);
                    String[] fields = line.Split(new char[] { '\t' });
                    Console.WriteLine("{0} ({1}m) - located in {2}",
                          fields[nameColumn], fields[elevationColumn], fields[countryColumn]);
                }
            }
            
        }
    }

}
