﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using System.IO;
using System.Net;
using System.Threading;
using System.Diagnostics;

using Microsoft.Win32;
using Microsoft.Win32.SafeHandles;
using System.Runtime.InteropServices;

namespace geonamesThreads
{
    /// <summary>
    /// Memory mapped conversion functions. This class is unsafe as in imports the Windows kernel DLL to open and map files.
    /// </summary>
    public unsafe class GeonamesSearch
    {
        const int MAX_NUMBER_COLUMNS = 30;  // Maximum number of columns (fields) allowed in a geobase line
        const byte lf = 0x0A, tab = 0x09, zero = 0x30;
        const int nameColumn = 1;
        const int countryColumn = 8;
        const int elevationColumn = 15;
        const int targetElevation = 8000;

        /// <summary>
        /// Perform multithreaded geonames file search.
        /// </summary>
        /// <param name="inFile">Input file path name</param>
        /// <param name="threadPool">false: Use threads without a thread pool; true: Use a threadc poool</param>
        /// <param name="display">Display the located objects if true.</param>
        /// <returns>0 - no error returns are currently defined.</returns>
        public static Int64 geonamesSearch(String inFile, uint nThreads, bool threadPool, bool display)
        {
            Int64 iPos = 0, nFound = 0;
            FileMm mappedFile = null;

            try
            {
                mappedFile = new FileMm(inFile);
                byte* pIn = mappedFile.pIn;

                UInt64 fileLength = mappedFile.FileSize;

                // Create the search thread arguments, then create the threads to convert file stripes concurrently
                geonamesArg[] threadArgs = (from n in Enumerable.Range(0, (int)nThreads) select new geonamesArg(pIn, (uint)n, nThreads, fileLength)).ToArray();

                if (threadPool)
                {
                    ManualResetEvent finishedEvent = new ManualResetEvent(false);
                    int remainingCallbacks = (int)nThreads;

                    // Create the pooled tasks and wait for them all to complete.
                    threadArgs.ForAll((ta) =>
                        ThreadPool.QueueUserWorkItem(delegate
                        {
                            GeonamesSearch.ReadAndFilter(ta);
                            if (Interlocked.Decrement(ref remainingCallbacks) == 0) { finishedEvent.Set(); }
                        })
                    );

                    // Wait for all pooled tasks to complete
                    finishedEvent.WaitOne();
                }
                else
                {
                    Thread[] geonamesThread = (from n in Enumerable.Range(0, (int)nThreads) select new Thread(new ParameterizedThreadStart(ReadAndFilter))).ToArray();
                    // Start all the threads. The static ForAll method is implemented at the end of the file.
                    threadArgs.ForAll((ta) => geonamesThread[ta.threadNumber].Start(ta));
                    threadArgs.ForAll((ta) => { geonamesThread[ta.threadNumber].Join(); });
                }

                // Accumulate the number of identified qualifying locations.
                // Merge results. Use just two fields (elevation and index) in QualifyingLocation objects.
                List<QualifyingLocation> ql = new List<QualifyingLocation>();
                threadArgs.ForAll((ta) =>
                {
                    for (int i = 0; i < ta.NumberLocations(); i++) { ql.Add(ta.GetLocation(i)); }
                });

                LocationComparer lc = new LocationComparer();
                ql.Sort(lc);
                nFound = ql.Count();
                System.Text.Encoding encoding = System.Text.UnicodeEncoding.UTF8;
                if (display) foreach (QualifyingLocation qLoc in ql)
                {
                    // The name field is UTF-8 and must be encoded from the byte[] array
                    byte[] name = new byte[qLoc.ColumnStart[nameColumn + 1] - qLoc.ColumnStart[nameColumn]];
                    int iName = 0;
                    for (UInt64 iScan = qLoc.ColumnStart[nameColumn]; pIn[iScan] != tab; iScan++, iName++) { name[iName] = pIn[iScan]; }
                    Console.Write("{0} ({1}m) - located in ", encoding.GetString(name), qLoc.Elevation);
                    for (UInt64 iScan = qLoc.ColumnStart[countryColumn]; pIn[iScan] != tab; iScan++) { Console.Write("{0}", (Char)pIn[iScan]); }
                    Console.WriteLine();
                }
            }
            catch (Exception e)
            {
                Console.Error.WriteLine("Error. geonamesMT. {0}.", e.Message);
                return iPos;
            }
            finally
            {
                mappedFile.Dispose();
            }

            return nFound;
        }

        /// <summary>
        /// ReadAndFilter thread/work item function. MemoryMapped version.
        /// </summary>
        /// <param name="cargObject"></param>
        /// <remarks>    /* Thread function to perform file filtering on one "partition" within the file
        ///     geonamesArg->threadNum is the "partition number" (0, 1, ..., nThreads)
        ///          Partition i processes text starting at position i * partitionSize on the 
        ///          file (which is mapped). It first identifies the start of the first full
        ///          text line within the partition, and it will process into the next partition, if
        ///          necessary, to find the end of a text line.
        ///          Text line records consist of tab-delimited fields </remarks>
        private static void ReadAndFilter(object geonamesArgObject)
        {
            geonamesArg geonamesArg = geonamesArgObject as geonamesArg;
            if (geonamesArg == null)
            {
                // This should never happen.
                Console.Error.WriteLine("Error. ReadAndFilter thread function. Argument object is not a geonamesArg.");
                return; // The error will be detected as a failure to convert all the bytes in the file.
            }

            uint iThread = geonamesArg.threadNumber;
            uint nThreads = geonamesArg.nThreads;
            UInt64 fileSize = geonamesArg.fileSize;
            UInt64 partitionSize = geonamesArg.partitionSize;
            UInt64 iPos = 0;
            byte* pFile = geonamesArg.pFile;
            FilteredElevations filteredElevations = geonamesArg.Elevations;

            UInt64 lineEndPos, curPos, elevationEndPos;
            UInt64[] columnStart = new UInt64[MAX_NUMBER_COLUMNS + 1];
            int columnNumber, elevation;  // Location of column start in currently scanned line

            /* Scan the full lines in this file partition (first new line after pFile+curPos) until the end of the partition, scanning
             * beyond the end of the partition for the full line.
             * Record each line with a sufficiently high elevation, retaining the name, country, and elevation - Assume that the number of
             * qualifying locations is small and does not exceed MAX_FILTERED_ELEVATION (additional qualifying elevations are ignored - an enhancement
             * would be to report an error). Records are maintained in "pResults"
             */

            try
            {
                curPos = partitionSize * iThread;
                if (curPos > 0)
                {
                    while (curPos < fileSize && pFile[curPos] != lf) { curPos++; }
                    curPos++; // Index of start of the line.
                }
                while (curPos < fileSize && curPos < partitionSize * (iThread + 1) && filteredElevations.NumberFound < filteredElevations.MaxSize)
                {
                    // (curPos == 0 && iThread == 0) || pFile[curPos-1] == '\n'
                    // Scan each line, recording the start of each tab-separated column in columnStart[]. Test the elevation and record as appropriate
                    // This code would be much simpler with a "Split" function (as in .NET and Perl). However, performance might suffer as you would
                    // need to allocate memory to store all the strings; we're only interested in the elevation at this point.
                    columnNumber = 0;
                    while (pFile[curPos] != lf && curPos < fileSize)
                    {
                        columnStart[columnNumber++] = curPos;
                        while (pFile[curPos] != tab && pFile[curPos] != lf) { curPos++; }
                        if (pFile[curPos] == tab) { curPos++; }
                    }
                    lineEndPos = curPos;
                    // The line scan is complete and field boundaries are marked in columnStart[]
                    // Extract the elevation, if any.
                    if (columnNumber < elevationColumn) continue;  // No elevation data in this line. Go to the next line.
                    elevationEndPos = (columnNumber == elevationColumn) ? lineEndPos - 1 : columnStart[elevationColumn + 1] - 1;
                    elevation = 0;

                    // Brute force ascii to decimal conversion to avoid modifying the file temporarily. The file and map would then need to be read/write 
                    // Many fields are empty.
                    for (iPos = columnStart[elevationColumn]; iPos < elevationEndPos; iPos++)
                    {
                        elevation = elevation * 10 + (int)pFile[iPos] - zero;
                    }
                    if (elevation >= targetElevation)
                    {
                        // Record this location
                        int index = filteredElevations.NumberFound;
                        QualifyingLocation ql = new QualifyingLocation(elevation, columnStart);
                        filteredElevations.SetQualifyingLocation(index, ql);
                        filteredElevations.NumberFound++;
                    }
                    curPos++;
                }

            }

            catch (Exception e)
            {
                throw new Exception("ReadAndFilter access exception. " + e.Message);
            }

            return;
        }

        /// <summary>
        /// Thread argument for the geonamesThreads threads and work items. MemoryMapped version.
        /// </summary>
        unsafe private class geonamesArg
        {
            const int MAX_FILTERED_ELEVATION = 64;  // Maximum number of records that a single thread can filter. We assume that the number of results is low (but detect a violation)
            public geonamesArg(byte* pFile, uint iThread, uint nThreads, UInt64 fileSize)
            {
                _threadNum = iThread;
                _nThreads = nThreads;
                _fileSize = fileSize;
                _pFile = pFile;
                _partitionSize = (fileSize + nThreads - 1) / nThreads;
                _filteredElevations = new FilteredElevations(MAX_FILTERED_ELEVATION);
            }

            private uint _threadNum;
            private uint _nThreads;
            private UInt64 _fileSize;
            private byte* _pFile;
            private UInt64 _partitionSize;
            private FilteredElevations _filteredElevations;


            public UInt64 fileSize { get { return _fileSize; } }
            public uint threadNumber { get { return _threadNum; } }
            public uint nThreads { get { return _nThreads; } }
            public byte* pFile { get { return _pFile; } }
            public UInt64 partitionSize { get { return _partitionSize; } }

            public int NumberLocations() { return _filteredElevations.NumberFound; }

            public QualifyingLocation GetLocation(int i)
            {
                return this._filteredElevations.Location[i];        //_qualifyingLocations[i];
            }

            public FilteredElevations Elevations { get { return _filteredElevations; } }

            public override String ToString()
            {
                return "InteropMemoryMappedGeonamesImplementation.geonamesArg. thread/task: " + _threadNum.ToString() + ", Length: " + _fileSize.ToString();
            }
        }

        private class QualifyingLocation
        {
            private int _elevation;
            private UInt64[] _columnStart;

            public QualifyingLocation(int el, UInt64[] cs)
            {
                _elevation = el;
                _columnStart = new UInt64[cs.Count()];
                for (int i = 0; i < cs.Count(); i++) { _columnStart[i] = cs[i]; }
            }

            public int Elevation { get { return _elevation; } set { _elevation = value; } }
            public UInt64[] ColumnStart { get { return _columnStart; } set { _columnStart = value; } }
        }

        private class LocationComparer : IComparer<QualifyingLocation>
        {
            public int Compare(QualifyingLocation q1, QualifyingLocation q2)
            {
                // Sort in descending order
                if (q1.Elevation < q2.Elevation) return +1;
                if (q1.Elevation > q2.Elevation) return -1;
                return 0;
            }
        }

        private class FilteredElevations
        {
            private QualifyingLocation[] _qualifyingLocations;
            private int _numberFound;
            private int _maxSize;

            public FilteredElevations(int max)
            {
                _numberFound = 0;
                _maxSize = max;
                _qualifyingLocations = new QualifyingLocation[max];
            }

            public void SetQualifyingLocation(int index, QualifyingLocation ql) { _qualifyingLocations[index] = new QualifyingLocation(ql.Elevation, ql.ColumnStart); }
            public int NumberFound { get { return _numberFound; } set { _numberFound = value; } }
            public int MaxSize { get { return _maxSize; } }
            public QualifyingLocation[] Location { get { return _qualifyingLocations; } }
        }

    }
}

/// <summary>
/// ForAll extension that applies to all IEnumerable types
/// </summary>
/// <remarks>Copied from page 52 of "Effective C#: 50 Specific Ways to Improve Your C#" by Bill Wagner, Addison-Wesley Effective Software Development Series, 2010.
/// This is used in the multithreaded and thread pool implementations to help replace explicit loops with query syntax.</remarks>
public static class Extensions
{
    public static void ForAll<T>(
        this IEnumerable<T> sequence,
        Action<T> action)
    {
        foreach (T item in sequence)
            action(item);
    }
}
