﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using System.IO;
using System.Net;
using System.Threading;
using System.Diagnostics;

using Microsoft.Win32;
using Microsoft.Win32.SafeHandles;
using System.Runtime.InteropServices;

namespace geonamesThreads
{
    /// <summary>
    /// Geonames test program using C#, .NET threads, and memory mapped files. Compare performance with the Native code version.
    /// </summary>
    /// <remarks>
    /// August 4, 2010. Johnson M. Hart (Copyright 2010 JMH Associates, Inc.)
    /// Background: 
    /// 1.  "LINQ to Objects Using C# 4.0" by Troy Magennis, Addison-Wesley, 2010. "LINQ Data Parallelism", pp 272ff.
    ///     This section describes a PLINQ solution to process an 825MB "geographical names" text file.
    /// 	This program will solve the same problem using native code and techniques from the next reference.
    /// 2.  "Windows Parallelism, Fast File Searching, and Speculative Processing" by Johnson M. Hart, June 30, 2010.
    ///     http://www.informit.com/articles/article.aspx?p=1606242.
    /// 	The project Projects2008_64\compMP (source file CHAPTR14\compMP.c) implements the the solution discussed in the article.
    /// 3.	"Sequential File Processing: A Fast Path from Native Code to .NET with Visits to Parallelism and Memory Mapped Files" by Johnson M. Hart, June 13, 2010.
    /// 	is a related C#/.NET solution (not using LINQ). The article might appear somewhere in the future. A current copy (article, code, etc.)
    /// 	is at http://www.jmhartsoftware.com/SequentialFileProcessingSupport.html.
    /// 4.  Windows System Programming Edition 4, Chapters 5, 7-10, and 14, describe the Windows API features that are used in this program.
    /// 
    /// The code here uses native code to solve the same problem that Magennis (#1) solves with PLINQ so as to compare performance.
    /// Comment: Assume that we want to optimize performance, even if the code is less elegant and more difficult to write.

    /// Performance Results Overview, elapsed time (seconds) running on a 4-CPU system. - TO BE UPDATED
    /// 	#CPUs		Native, 64-bit code.		Native, 32-bit code.		PLINQ (#1 above), 32-bit code
    /// 		1			8.613						7.678						28.569
    /// 		2			3.570						3.398						14.434
    /// 		4			1.937						1.845						10.630
    /// 
    /// Quick conclusions:
    /// 	1.  The native code solution with memory mapped files is considrably faster than the PLINQ solution, but a factor of ~4
    /// 	2.  The PLINQ solution does not scale with the number of processors beyond 2 processors. The native code solution does scale well.
    /// 	3.  The native code, however, is far more complex and much less elegant than the PLINQ code. This complexity, in turn, can lead
    /// 		to maintenance difficulties and unpleasant bugs. The performance, however, may be well worth the trouble.
    /// 	4.  32-bit code is marginally faster than 64-bit code, but larger files than used in this example will require
    /// 		64-bit code.
    /// 	5.  A faster PLINQ solution may be possible, however, with a Memory Mapped File class that supports the IEnumberable interface.
    /// 		That will be a future project. In fact, the geonamesPLINQ project implements a fast solution, which is described
    /// 		in http://msdn.microsoft.com/en-us/magazine/gg535673.aspx.
    ///     6.  The genamesParallel project shows how to use .NET 4.0 and the TPL's Parallel class 
    /// 	7.  For the time being, it's safe to say that native code using threads and memory mapped files can provide excellent performance
    /// 		for numerous file processing problems, and the same is true for .NET threads. Furthermore, the basic coding pattern is well established with the WSP4 examples 
    /// 		and Background articles #2 and #3. Nonetheless, the code can become complex, and the .NET solutions are much easier and
    /// 		more elegant.
    /// 
    ///   Problem overview: Find all locations with an elevation (column 15) > 8000 meters. Retain the name (column 1), the country (column 8)
    /// and the elevation  Sort in descenting order. Display the results (there are 16 places above 8000 meters. Only two names were familiar to me).
    /// 
    ///  geonames file [ nThreads ] [ display ]
    ///        file: pathname for the geonames database text file.
    /// 			Download and unzip from: http://download.geonames.org/export/dump/allCountries.zip. The unzipped file, allCountries.txt, is about 825MB.
    /// 			The fields are tab-separated, with one geographical location per line.
    /// 			As of Dec 5, 2010, there is at least one incorrect elevation, for "Señal Salumpite  (8600m) - located in PE". This is clearly a mistake, but is not relevant to this example.
    /// 			I can't vouch for any of the data in this file, so don't plan your next vacation based on this information!
    ///        nThreads: Number of independent threads (parallel work items to process the file)
    ///                  Default: Number of processors on this system.
    /// 				 POTENTIAL ENHANCEMENT: If the file size is too large compared to physical memory, use the conventional]
    /// 					file processing solution. See "Windows Parallelism, Fast File Searching, and Speculative Processing"
    /// 					for more information.
    /// 				 ADDITIONAL ENHANCEMENT. Try thread pooling
    /// 		display: If 0, don't display the records. If omitted, or non-zero, do display.
    /// 
    ///  Comment: This program may fail for large files, especially on 32-bit systems
    ///  where you have, at most, 3 GB to work with (of course, you have much less
    ///  in reality, and you need to map the file). However, the example file is small
    ///  enough that it should work on any system with 2GB RAM. 
    ///  </remarks>

    class Program
    {
        private static uint nThreads = (uint)Environment.ProcessorCount;       // Default: Maximum number of threads or work items running concurrently.

        static void Main(string[] args)
        {
            String[] concurrencyTechniques = { "MultiThreaded", "ThreadPool" };   // Default: None

            if (args.Length < 1)
            {
                Console.Error.WriteLine("Usage Error: geonamesThreads inFile [ nThreads [ threadingTechnique ] ]");
                return;
            }

            String inFile = args[0];
            if (args.Length >= 2) nThreads = (uint)int.Parse(args[1]);
            String threadingTechnique = "MultiThreaded";
            if (args.Length >= 3) threadingTechnique = args[2];

            if (!concurrencyTechniques.Contains(threadingTechnique))
            {
                Console.Error.WriteLine("Error: Threading technique {0} is not supported.", threadingTechnique);
                return;
            }

            bool display = args.Length < 4 || (args.Length >= 4 && int.Parse(args[3]) > 0);
            Console.WriteLine("GeonamesThreads - Data file: {0}, nThreads: {1}, Threading Technique: {2}.", inFile, nThreads, threadingTechnique);

            Int64 startTime = DateTime.Now.Ticks;
            Int64 nFound = 0;
            try
            {
                if (threadingTechnique.Equals("MultiThreaded"))
                {
                    nFound = GeonamesSearch.geonamesSearch(inFile, nThreads, false, display);
                }
                else if (threadingTechnique.Equals("ThreadPool"))
                {
                    nFound = GeonamesSearch.geonamesSearch(inFile, nThreads, true, display);
                }
                else
                {
                    Console.Error.WriteLine("Error. This threading option is not supported.");
                }
            }

            catch (Exception e)
            {
                Console.Error.WriteLine("Error. Geonames file search failed. {0}", e.Message);
                return;
            }


            Int64 elapsedTime = DateTime.Now.Ticks - startTime;

            Console.WriteLine("Query Count: {0}. Elapsed Query time: {1}.{2} seconds.", nFound, elapsedTime / 10000000, (elapsedTime % 10000000) / 10000);
            return;
        }
    }
}